"""High-level interface to Structure Function.

.. todo::
    refer to the sf-esf overview

"""

import logging

from .esf import esf, tmc

logger = logging.getLogger(__name__)


class StructureFunction:
    r"""Represent an abstract structure function.

    This class acts as an intermediate handler between the :py:class:`Runner`
    exposed to the outside and the :py:class:`EvaluatedStructureFunction`
    which compute the actual observable.

    Parameters
    ----------
    obs_name : ObservableName
        name
    runner : yadism.runner.Runner
        parent reference

    """

    def __init__(self, obs_name, runner):
        # internal managers
        self.obs_name = obs_name
        self.runner = runner
        self.esfs = []
        self.cache = {}
        logger.debug("Init %s", self)

    def __repr__(self):
        return self.obs_name.name

    def __len__(self):
        return len(self.esfs)

    def load(self, kinematic_configs):
        r"""Load all kinematic configurations from the run card.

        Parameters
        ----------
            kinematic_configs : list(dict)
                run card input

        """
        self.esfs = []
        # iterate F* configurations
        for kinematics in kinematic_configs:
            self.esfs.append(self.get_esf(self.obs_name, kinematics, use_raw=False))

    def get_esf(self, obs_name, kinematics, *args, use_raw=True, force_local=False):
        """Return a :py:class:`EvaluatedStructureFunction` instance.

        This wrappers allows

        - TMC to to access raw computations
        - heavy quark matching schemes to access their light counter parts

        It also implements an internal caching system, to speed up the integrals
        in TMC.

        Parameters
        ----------
        obs_name : .observable_name.ObservableName
            structure function name
        kinematics : dict
            kinematic configuration
        args : any
            further arguments passed down to the instance
        use_raw : bool
            eventually use the ESFTMC? (or just the uncorrected one)

        Returns
        -------
        obj : EvaluatedStructureFunction
            created object

        """
        # TODO rethink and refactor method - only used by TMC
        # TODO remove force_local
        # if force_local is active suppress caching to avoid circular dependecy
        if force_local:
            obj = esf.EvaluatedStructureFunction(
                kinematics, self.obs_name, self.runner.configs
            )
            return obj
        # else we're happy to cache
        # is it us or do we need to delegate?
        if obs_name == self.obs_name:
            # convert to tuple
            key = list(kinematics.values())
            use_tmc_if_available = not use_raw and self.runner.configs.TMC != 0
            key.append(use_tmc_if_available)
            key = tuple(key)
            # TODO how to incorporate args?
            # search
            try:
                return self.cache[key]
            except KeyError:
                if use_tmc_if_available:
                    obj = tmc.ESFTMCmap[obs_name.kind](self, kinematics)
                else:
                    obj = esf.EvaluatedStructureFunction(
                        kinematics, self.obs_name, self.runner.configs
                    )
                self.cache[key] = obj
                return obj
        else:
            # ask our parent (as always)
            return self.runner.get_sf(obs_name).get_esf(obs_name, kinematics, *args)

    @property
    def elements(self):
        """Collect the computed observables."""
        return self.esfs

    def get_result(self):
        r"""Collect the results from all childrens.

        Returns
        -------
        output : list(ESFResult)
            all children outputs

        """
        return list(elem.get_result() for elem in self.elements)

    def drop_cache(self):
        """Drop temporary cache.

        This preserves final results, since they are not part of the cache.

        """
        self.cache = {}



import numpy as np
from eko import constants

from ..partonic_channel import RSL
from . import partonic_channel as pc
from . import raw_cc


class Rplus(pc.ChargedCurrentBase):
    def __init__(self, *args, m1sq):
        super().__init__(*args, m1sq=m1sq)
        self.lo = 1.0

    def LO(self):
        return RSL.from_delta(self.lo)

    def NLO(self):
        norm = constants.CF
        lnomx = raw_cc.lnomx * norm * self.lo
        omx = raw_cc.omx(self.y) * norm * self.lo
        delta = raw_cc.f3sv(self.y) * norm

        def reg(z, _args):
            return norm * raw_cc.f3r(self.y, z) - (lnomx * np.log(1.0 - z) + omx) / (
                1.0 - z
            )

        return RSL.from_distr_coeffs(reg, (delta, omx, lnomx))



import LeProHQ
import numpy as np

from ..partonic_channel import RSL
from . import partonic_channel as pc


class GluonVV(pc.NeutralCurrentBase):
    def NLO(self):
        """
        |ref| implements :eqref:`D.3`, :cite:`felix-thesis`.
        """

        def cg(z, _args):
            if self.is_below_pair_threshold(z):
                return 0.0
            return (
                self._FHprefactor
                / z
                * LeProHQ.cg0("x2g1", "VV", self._xi, self._eta(z))
            )

        return RSL(cg)

    def NNLO(self):
        """
        |ref| implements |NNLO| (heavy) gluon coefficient function,
        :eqref:`D.14` of :cite:`felix-thesis`.
        """

        def cg(z, _args):
            if self.is_below_pair_threshold(z):
                return 0.0
            return (
                self._FHprefactor
                / z
                * (4.0 * np.pi) ** 2
                * (
                    LeProHQ.cg1("x2g1", "VV", self._xi, self._eta(z))
                    + LeProHQ.cgBar1("x2g1", "VV", self._xi, self._eta(z))
                    * np.log(self._xi)
                )
            )

        return RSL(cg)


class GluonAA(GluonVV):
    def NLO(self):
        """
        |ref| implements :eqref:`D.6`, :cite:`felix-thesis`.
        """

        def cg(z, _args):
            if self.is_below_pair_threshold(z):
                return 0.0
            return (
                self._FHprefactor
                / z
                * LeProHQ.cg0("x2g1", "AA", self._xi, self._eta(z))
            )

        return RSL(cg)

    def NNLO(self):
        """
        |ref| implements |NNLO| (heavy) gluon coefficient function,
        :eqref:`D.14` of :cite:`felix-thesis`.
        """

        def cg(z, _args):
            if self.is_below_pair_threshold(z):
                return 0.0
            return (
                self._FHprefactor
                / z
                * (4.0 * np.pi) ** 2
                * (
                    LeProHQ.cg1("x2g1", "AA", self._xi, self._eta(z))
                    + LeProHQ.cgBar1("x2g1", "AA", self._xi, self._eta(z))
                    * np.log(self._xi)
                )
            )

        return RSL(cg)


class SingletVV(pc.NeutralCurrentBase):
    def NNLO(self):
        """
        |ref| implements |NNLO| (heavy) singlet coefficient function,
        :eqref:`D.43` of :cite:`felix-thesis`.
        """

        def cq(z, _args):
            if self.is_below_pair_threshold(z):
                return 0.0
            return (
                self._FHprefactor
                / z
                * (4.0 * np.pi) ** 2
                * (
                    LeProHQ.cq1("x2g1", "VV", self._xi, self._eta(z))
                    + LeProHQ.cqBarF1("x2g1", "VV", self._xi, self._eta(z))
                    * np.log(self._xi)
                )
            )

        return RSL(cq)


class SingletAA(pc.NeutralCurrentBase):
    def NNLO(self):
        """
        |ref| implements |NNLO| (heavy) singlet coefficient function,
        :eqref:`D.43` of :cite:`felix-thesis`.
        """

        def cq(z, _args):
            if self.is_below_pair_threshold(z):
                return 0.0
            return (
                self._FHprefactor
                / z
                * (4.0 * np.pi) ** 2
                * (
                    LeProHQ.cq1("x2g1", "AA", self._xi, self._eta(z))
                    + LeProHQ.cqBarF1("x2g1", "AA", self._xi, self._eta(z))
                    * np.log(self._xi)
                )
            )

        return RSL(cq)


class NonSinglet(pc.NeutralCurrentBase):
    def NNLO(self):
        """
        |ref| implements |NNLO| (heavy) NS coefficient function,
        :eqref:`D.64` of :cite:`felix-thesis`.
        """

        def dq(z, _args):
            if self.is_below_pair_threshold(z):
                return 0.0
            return (
                self._FHprefactor
                / z
                * (4.0 * np.pi) ** 2
                * (LeProHQ.dq1("x2g1", "VV", self._xi, self._eta(z)))
            )

        def Adler(_x, _args):
            # add minus sign
            return -LeProHQ.Adler("x2g1", "VV", self._xi)

        return RSL(dq, loc=Adler)



def W2(x, Q2, MP=0.0):
    """Mass squared :math:`W^2` of the proton-boson system.

    In particular, it is the mass squared of the system X recoiling against the
    scattered lepton :cite:`Zyla:2020zbs`.

    Parameters
    ----------
    x: float
        Bjorken x
    Q2: float
        photon virtuality
    MP: float
        mass of the proton (default: ``0.``)

    Returns
    -------
    float
        :math:`W^2` value computed

    """
    return MP**2 + Q2 * (1 - x) / x



from ..partonic_channel import RSL
from . import fl_nc, n3lo, nnlo


class NonSingletEven(fl_nc.NonSinglet):
    pass


class Gluon(fl_nc.Gluon):
    pass


class Singlet(fl_nc.Singlet):
    pass


class NonSingletOdd(fl_nc.NonSinglet):
    def NNLO(self):
        """
        |ref| implements :eqref:`2.9`, :cite:`vogt-f2lcc`.
        """

        return RSL(
            nnlo.xclns2p.clnc2a, loc=nnlo.xclns2p.clnc2c, args=dict(reg=[self.nf])
        )

    def N3LO(self):
        """
        |ref| implements the difference between :eqref:`2.10`, and :eqref:`3.7`, :cite:`Davies:2016ruz`
        """
        return RSL(
            n3lo.xclns3p.clnm3a_fl2,
            loc=n3lo.xclns3p.clnm3c_fl2,
            args=dict(reg=[self.nf]),
        )



import adani

from ..light import f2_nc as light
from ..partonic_channel import RSL
from . import partonic_channel as pc
from . import raw_nc


class AsyGluon(pc.NeutralCurrentBaseAsy):
    hs3 = adani.HighScaleSplitLogs(3, "2", "g", "gm")


class AsySinglet(pc.NeutralCurrentBaseAsy):
    hs3 = adani.HighScaleSplitLogs(3, "2", "q", "exact")


class AsyLLGluon(AsyGluon):
    def NLO(self):
        def cg_LL_NLO(z, args):
            L = args[0]
            return raw_nc.c2g1am0_aq(z) * L

        return RSL(cg_LL_NLO, args=[self.L])

    def NNLO(self):
        def cg_LL_NNLO(z, args):
            L = args[0]
            return (raw_nc.c2g2am0_aq2(z) - raw_nc.c2g2am0_aqf(z)) * L**2

        return RSL(cg_LL_NNLO, args=[self.L])

    def N3LO(self):
        def cg_LL_N3LO(z, args):
            L = -args[0]
            nf = int(args[1])
            return self.hs3.LL(z, nf) * L**3

        return RSL(cg_LL_N3LO, args=[self.L, self.nf])


class AsyNLLGluon(AsyGluon):
    def NLO(self):
        def cg_NLL_NLO(z, _args):
            return raw_nc.c2g1am0_a0(z)

        return RSL(cg_NLL_NLO, args=[self.L])

    def NNLO(self):
        def cg_NLL_NNLO(z, args):
            L = args[0]
            return (raw_nc.c2g2am0_aq(z) - raw_nc.c2g2am0_af(z)) * L

        return RSL(cg_NLL_NNLO, args=[self.L])

    def N3LO(self):
        def cg_NLL_N3LO(z, args):
            L = -args[0]
            nf = int(args[1])
            return self.hs3.NLL(z, nf) * L**2

        return RSL(cg_NLL_N3LO, args=[self.L, self.nf])


class AsyNNLLGluon(AsyGluon):
    def NNLO(self):
        def cg_NNLL_NNLO(z, _args):
            return raw_nc.c2g2am0_a0(z)

        return RSL(cg_NNLL_NNLO)

    def N3LO(self):
        def cg_NNLL_N3LO(z, args):
            L = -args[0]
            nf = int(args[1])
            return self.hs3.N2LL(z, nf) * L

        return RSL(cg_NNLL_N3LO, args=[self.L, self.nf])


class AsyNNNLLGluon(AsyGluon):
    variation_map = {-1: 2, 0: 0, 1: 1}

    def N3LO(self):
        def cg_NNNLL_N3LO(z, args):
            nf = int(args[0])
            variation = int(args[1])
            return self.hs3.N3LL(z, nf).ToVect()[self.variation_map[variation]]

        return RSL(cg_NNNLL_N3LO, args=[self.nf, self.n3lo_cf_variation])


class AsyLLSinglet(AsySinglet):
    def NNLO(self):
        def cps_LL_NNLO(z, args):
            L = args[0]
            return (raw_nc.c2ps2am0_aq2(z) - raw_nc.c2ps2am0_aqf(z)) * L**2

        return RSL(cps_LL_NNLO, args=[self.L])

    def N3LO(self):
        def cps_LL_N3LO(z, args):
            L = -args[0]
            nf = int(args[1])
            return self.hs3.LL(z, nf) * L**3

        return RSL(cps_LL_N3LO, args=[self.L, self.nf])


class AsyNLLSinglet(AsySinglet):
    def NNLO(self):
        def cps_NLL_NNLO(z, args):
            L = args[0]
            return (raw_nc.c2ps2am0_aq(z) - raw_nc.c2ps2am0_af(z)) * L

        return RSL(cps_NLL_NNLO, args=[self.L])

    def N3LO(self):
        def cps_NLL_N3LO(z, args):
            L = -args[0]
            nf = int(args[1])
            return self.hs3.NLL(z, nf) * L**2

        return RSL(cps_NLL_N3LO, args=[self.L, self.nf])


class AsyNNLLSinglet(AsySinglet):
    def NNLO(self):
        def cps_NNLL_NNLO(z, _args):
            return raw_nc.c2ps2am0_a0(z)

        return RSL(cps_NNLL_NNLO)

    def N3LO(self):
        def cps_NNLL_N3LO(z, args):
            L = -args[0]
            nf = int(args[1])
            return self.hs3.N2LL(z, nf) * L

        return RSL(cps_NNLL_N3LO, args=[self.L, self.nf])


class AsyNNNLLSinglet(AsySinglet):
    def N3LO(self):
        def cps_NNNLL_N3LO(z, args):
            nf = int(args[0])
            return self.hs3.N3LL(z, nf).GetCentral()

        return RSL(cps_NNNLL_N3LO, args=[self.nf])


class AsyLLNonSinglet(pc.NeutralCurrentBaseAsy):
    def NNLO(self):
        def cns_LL_NNLO_reg(z, args):
            L = args[0]
            return raw_nc.c2ns2am0_aq2(z) * L**2

        def cns_LL_NNLO_sing(z, args):
            L = args[0]
            return raw_nc.c2ns2bm0_aq2(z) * L**2

        def cns_LL_NNLO_loc(z, args):
            L = args[0]
            return raw_nc.c2ns2cm0_aq2(z) * L**2

        return RSL(cns_LL_NNLO_reg, cns_LL_NNLO_sing, cns_LL_NNLO_loc, args=[self.L])


class AsyNLLNonSinglet(pc.NeutralCurrentBaseAsy):
    def NNLO(self):
        def cns_NLL_NNLO_reg(z, args):
            L = args[0]
            return raw_nc.c2ns2am0_aq(z) * L

        def cns_NLL_NNLO_sing(z, args):
            L = args[0]
            return raw_nc.c2ns2bm0_aq(z) * L

        def cns_NLL_NNLO_loc(z, args):
            L = args[0]
            return raw_nc.c2ns2cm0_aq(z) * L

        return RSL(cns_NLL_NNLO_reg, cns_NLL_NNLO_sing, cns_NLL_NNLO_loc, args=[self.L])


class AsyNNLLNonSinglet(pc.NeutralCurrentBaseAsy):
    def NNLO(self):
        def cns_NNLL_NNLO_reg(z, _args):
            return raw_nc.c2ns2am0_a0(z)

        def cns_NNLL_NNLO_sing(z, _args):
            return raw_nc.c2ns2bm0_a0(z)

        def cns_NNLL_NNLO_loc(z, _args):
            return raw_nc.c2ns2cm0_a0(z)

        return RSL(cns_NNLL_NNLO_reg, cns_NNLL_NNLO_sing, cns_NNLL_NNLO_loc)


class AsyNNNLLNonSinglet(pc.NeutralCurrentBaseAsy):
    pass


class AsyLLIntrinsic(pc.PartonicChannelAsyLLIntrinsic):
    light_cls = light.NonSinglet


class AsyNLLIntrinsicMatching(pc.PartonicChannelAsyNLLIntrinsicMatching):
    light_cls = light.NonSinglet


class AsyNLLIntrinsicLight(pc.PartonicChannelAsyNLLIntrinsicLight):
    light_cls = light.NonSinglet



from ..partonic_channel import EmptyPartonicChannel


class AsyLLNonSinglet(EmptyPartonicChannel):
    pass


class AsyNLLNonSinglet(EmptyPartonicChannel):
    pass


class AsyNNLLNonSinglet(EmptyPartonicChannel):
    pass


class AsyNNNLLNonSinglet(EmptyPartonicChannel):
    pass


class AsyLLIntrinsic(EmptyPartonicChannel):
    pass


class AsyNLLIntrinsicMatching(EmptyPartonicChannel):
    pass


class AsyNLLIntrinsicLight(EmptyPartonicChannel):
    pass



import adani

from ..light import fl_nc as light
from ..partonic_channel import RSL, EmptyPartonicChannel
from . import partonic_channel as pc
from . import raw_nc


class AsyLLGluon(EmptyPartonicChannel):
    pass


class AsyGluon(pc.NeutralCurrentBaseAsy):
    hs3 = adani.HighScaleSplitLogs(3, "L", "g", "gm")


class AsySinglet(pc.NeutralCurrentBaseAsy):
    hs3 = adani.HighScaleSplitLogs(3, "L", "q", "exact")


class AsyNLLGluon(AsyGluon):
    def NLO(self):
        def cg_NLL_NLO(z, _args):
            return raw_nc.clg1am0_a0(z)

        return RSL(cg_NLL_NLO, args=[self.L])

    def NNLO(self):
        def cg_NLL_NNLO(z, args):
            L = args[0]
            return (raw_nc.clg2am0_aq(z) - raw_nc.clg2am0_af(z)) * L

        return RSL(cg_NLL_NNLO, args=[self.L])

    def N3LO(self):
        def cg_NLL_N3LO(z, args):
            L = -args[0]
            # this term does not depend on nf so setting it to zero
            return self.hs3.NLL(z, 0) * L**2

        return RSL(cg_NLL_N3LO, args=[self.L])


class AsyNNLLGluon(AsyGluon):
    def NNLO(self):
        def cg_NNLL_NNLO(z, _args):
            return raw_nc.clg2am0_a0(z)

        return RSL(cg_NNLL_NNLO)

    def N3LO(self):
        def cg_NNLL_N3LO(z, args):
            L = -args[0]
            nf = int(args[1])
            return self.hs3.N2LL(z, nf) * L

        return RSL(cg_NNLL_N3LO, args=[self.L, self.nf])


class AsyNNNLLGluon(AsyGluon):
    def N3LO(self):
        def cg_NNNLL_N3LO(z, args):
            nf = int(args[0])
            return self.hs3.N3LL(z, nf).GetCentral()

        return RSL(cg_NNNLL_N3LO, args=[self.nf])


class AsyLLSinglet(EmptyPartonicChannel):
    pass


class AsyNLLSinglet(AsySinglet):
    def NNLO(self):
        def cps_NLL_NNLO(z, args):
            L = args[0]
            return (raw_nc.clps2am0_aq(z) - raw_nc.clps2am0_af(z)) * L

        return RSL(cps_NLL_NNLO, args=[self.L])

    def N3LO(self):
        def cps_NLL_N3LO(z, args):
            L = -args[0]
            # this term does not depend on nf so setting it to zero
            return self.hs3.NLL(z, 0) * L**2

        return RSL(cps_NLL_N3LO, args=[self.L])


class AsyNNLLSinglet(AsySinglet):
    def NNLO(self):
        def cps_NNLL_NNLO(z, _args):
            return raw_nc.clps2am0_a0(z)

        return RSL(cps_NNLL_NNLO)

    def N3LO(self):
        def cps_NNLL_N3LO(z, args):
            L = -args[0]
            # this term does not depend on nf so setting it to zero
            return self.hs3.N2LL(z, 0) * L

        return RSL(cps_NNLL_N3LO, args=[self.L])


class AsyNNNLLSinglet(AsySinglet):
    def N3LO(self):
        def cps_NNNLL_N3LO(z, args):
            nf = int(args[0])
            return self.hs3.N3LL(z, nf).GetCentral()

        return RSL(cps_NNNLL_N3LO, args=[self.nf])


class AsyLLNonSinglet(EmptyPartonicChannel):
    pass


class AsyNLLNonSinglet(pc.NeutralCurrentBaseAsy):
    def NNLO(self):
        def cns_NLL_NNLO(z, args):
            L = args[0]
            return raw_nc.clns2am0_aq(z) * L

        return RSL(cns_NLL_NNLO, args=[self.L])


class AsyNNLLNonSinglet(pc.NeutralCurrentBaseAsy):
    def NNLO(self):
        def cns_NNLL_NNLO(z, _args):
            return raw_nc.clns2am0_a0(z)

        return RSL(cns_NNLL_NNLO)


class AsyNNNLLNonSinglet(EmptyPartonicChannel):
    pass


class AsyLLIntrinsic(pc.PartonicChannelAsyLLIntrinsic):
    light_cls = light.NonSinglet


class AsyNLLIntrinsicMatching(pc.PartonicChannelAsyNLLIntrinsicMatching):
    light_cls = light.NonSinglet


class AsyNLLIntrinsicLight(pc.PartonicChannelAsyNLLIntrinsicLight):
    light_cls = light.NonSinglet



"""See :mod:`f3_cc` docstring for the name conventions."""

from .. import partonic_channel as epc
from ..partonic_channel import RSL
from . import f2_nc, n3lo, nlo, nnlo
from . import partonic_channel as pc


class NonSinglet(f2_nc.NonSinglet):
    @staticmethod
    def NLO():
        """|ref| implements :eqref:`155`, :cite:`moch-f3nc`."""

        return RSL.from_distr_coeffs(
            nlo.f3.ns_reg, (nlo.f2.ns_delta, nlo.f2.ns_omx, nlo.f2.ns_logomx)
        )

    def NNLO(self):
        """|ref| implements :eqref:`3.6` :cite:`vogt-f3cc`, or :eqref:`208`, :cite:`moch-f3nc`, or :eqref:`2.8` :cite:`Davies:2016ruz`."""

        return RSL(
            nnlo.xc3ns2p.c3nm2a, nnlo.xc3ns2p.c3ns2b, nnlo.xc3ns2p.c3nm2c, [self.nf]
        )

    def N3LO(self):
        """|ref| implements :eqref:`3.7` :cite:`vogt-f3cc`, or :eqref:`2.11` :cite:`Davies:2016ruz`."""
        return RSL(
            n3lo.xc3ns3p.c3nm3a,
            n3lo.xc3ns3p.c3ns3b,
            n3lo.xc3ns3p.c3nm3c,
            [self.nf],
        )


class Gluon(epc.EmptyPartonicChannel):
    pass


class Singlet(epc.EmptyPartonicChannel):
    pass


class Valence(pc.LightBase):

    def N3LO(self):
        """Part proportional to :math:`fl_{02}` of :eqref:`2.11` :cite:`Davies:2016ruz`."""

        return RSL(
            n3lo.xc3ns3p.c3nsv3a,
            args=[self.nf],
        )



"""Benchmark to QCDNUM."""

import pathlib

import numpy as np

from yadism import observable_name as on
from yadism.coefficient_functions.coupling_constants import CouplingConstants


# setup external PDF
class PdfCallable:
    """
    Wrapper to introduce lhapdf under QCDNUM.

    Parameters
    ----------
        pdf : lhapdf_like
            PDF set
    """

    def __init__(self, pdf):
        self.pdf = pdf

    def __call__(self, ipdf, x, qmu2, first):
        """
        Functor function.

        Parameters
        ----------
            ipdf : int
                pid
            x : float
                momentum fraction
            qmu2 : float
                momentum transfer

        Returns
        -------
            pdf(x)
        """
        if -6 <= ipdf <= 6:
            a = self.pdf.xfxQ2(ipdf, x, qmu2)
            return a
        return 0.0


def compute_qcdnum_data(
    theory, observables, pdf
):  #  pylint: disable=too-many-statements,too-many-branches,too-many-locals
    """
    Run QCDNUM to compute observables.

    Parameters
    ----------
        theory : dict
            theory runcard
        observables : dict
            observables runcard
        pdf : lhapdf_like
            PDF set

    Returns
    -------
        num_tab : dict
            QCDNUM numbers
    """
    if observables["prDIS"] == "CC":
        raise NotImplementedError("Charged current not supported in QCDNUM")

    import QCDNUM  # pylint:disable=import-outside-toplevel,import-error

    # remove QCDNUM cache files
    wname = "unpolarised-py.wgt"
    zmname = "zmstf-py.wgt"
    hqname = "hqstf-py.wgt"
    for f in [wname, zmname, hqname]:
        pathlib.Path(f).unlink(missing_ok=True)

    # init
    QCDNUM.qcinit(6, " ")

    # set params
    QCDNUM.setalf(theory["alphas"], theory["Qref"] ** 2)

    # make x and Q grids
    xmin = 0.00001
    q2min = 4
    q2max = 40
    for obs_name in observables["observables"]:
        # if not on.ObservableName.is_valid(obs):
        #    continue
        obs = on.ObservableName(obs_name)
        for kin in observables["observables"].get(obs_name, []):
            xmin = min(xmin, 0.5 * kin["x"])
            q2min = min(q2min, kin["Q2"])
            q2max = max(q2max, kin["Q2"])
    xarr = [xmin, np.power(xmin, 2.0 / 3.0)]
    xwarr = [1, 1]
    if xmin < 0.1:
        xarr += [0.1, 0.5]
        xwarr += [1, 1]
    iosp = 3
    n_x = 289
    n_q = 101
    af = 1.0 / theory["XIF"] ** 2
    bf = 0.0
    QCDNUM.gxmake(
        xarr, xwarr, n_x, iosp
    )  # grid walls, grid weights, points, interpolation type
    qarr = [q2min / af * 0.99, q2max / af * 1.01]
    qwarr = [1, 1]
    QCDNUM.gqmake(qarr, qwarr, n_q)

    # setup FNS
    mc = theory["mc"]
    mb = theory["mb"]
    mt = theory["mt"]
    iqc = QCDNUM.iqfrmq(mc**2)
    iqb = QCDNUM.iqfrmq(mb**2)
    iqt = QCDNUM.iqfrmq(mt**2)

    if theory["FNS"] == "FFNS":
        nfix = theory["NfFF"]
    else:
        nfix = 0
    QCDNUM.setcbt(nfix, iqc, iqb, iqt)

    # Try to read the weight file and create one if that fails
    QCDNUM.wtfile(1, wname)

    iset = 1

    # Try to read the ZM-weight file and create one if that fails
    if on.ObservableName.has_lights(observables["observables"].keys()):
        zmlunw = QCDNUM.nxtlun(10)
        _nwords, ierr = QCDNUM.zmreadw(zmlunw, zmname)
        if ierr != 0:
            QCDNUM.zmfillw()
            QCDNUM.zmdumpw(zmlunw, zmname)
        # set fact. scale
        QCDNUM.zmdefq2(af, bf)
        QCDNUM.zswitch(iset)

    # Try to read the HQ-weight file and create one if that fails
    if on.ObservableName.has_heavies(observables["observables"].keys()):
        hqlunw = QCDNUM.nxtlun(10)
        _nwords, ierr = QCDNUM.hqreadw(hqlunw, hqname)
        if ierr != 0:
            QCDNUM.hqfillw(3, [mc, mb, mt], af, bf)
            QCDNUM.hqdumpw(hqlunw, hqname)
        QCDNUM.hswitch(iset)

    # set ren scale
    arf = theory["XIR"] ** 2 / theory["XIF"] ** 2
    brf = 0
    QCDNUM.setabr(arf, brf)

    # func, pdf set number, nr. extra pdfs, thershold offset
    QCDNUM.extpdf(PdfCallable(pdf), iset, 0, 0)

    coupling = CouplingConstants.from_dict(theory, observables)
    num_tab = {}
    for obs_name in observables["observables"]:
        # if not on.ObservableName.is_valid(obs):
        #    continue
        obs = on.ObservableName(obs_name)
        kind_key = None
        if obs.kind == "F2":
            kind_key = 2
        elif obs.kind == "FL":
            kind_key = 1
        elif obs.name == "F3_light":
            kind_key = 3
        else:
            raise NotImplementedError(f"kind {obs.name} is not implemented!")

        q2s = []
        f_out = []

        # collect q2s
        for kin in observables["observables"].get(obs_name, []):
            if kin["Q2"] not in q2s:
                q2s.append(kin["Q2"])

        # loop over points
        for q2 in q2s:
            xs = []
            fs = []

            # get all the x corresponding to q2
            for kin in observables["observables"].get(obs_name, []):
                if kin["Q2"] == q2:
                    xs.append(kin["x"])
            # Use yadism to get all the weights
            weights = []
            for pid in range(-6, 7):
                if pid == 0:
                    pid = 21
                # F3
                if kind_key == 3:
                    w = np.sign(pid) * (
                        coupling.get_weight(pid, q2, "VA")
                        + coupling.get_weight(pid, q2, "AV")
                    )
                # F2 and FL
                else:
                    w = coupling.get_weight(pid, q2, "VV")
                    if observables["prDIS"] == "NC":
                        w += coupling.get_weight(pid, q2, "AA")
                weights.append(w)

            Q2s = [q2] * len(xs)
            # select fnc by flavor
            if obs.flavor == "light":
                QCDNUM.setord(1 + theory["PTO"])  # 1 = LO, ...
                fs.extend(QCDNUM.zmstfun(kind_key, weights, xs, Q2s, 1))

            elif obs.is_raw_heavy:
                # for HQ pto is not absolute but rather relative,
                # i.e., 1 loop DIS here meas LO[QCDNUM]
                if theory["PTO"] == 0:
                    fs = [0.0] * len(xs)
                else:
                    QCDNUM.setord(theory["PTO"])  # 1 = LO, ...

                if obs.flavor == "charm":
                    fs.extend(QCDNUM.hqstfun(kind_key, 1, weights, xs, Q2s, 1))
                elif obs.flavor == "bottom":
                    fs.extend(QCDNUM.hqstfun(kind_key, -2, weights, xs, Q2s, 1))
                elif obs.flavor == "top":
                    fs.extend(QCDNUM.hqstfun(kind_key, -3, weights, xs, Q2s, 1))
            else:
                raise NotImplementedError(f"flavor {obs.flavor} is not implemented!")

            # reshuffle output
            for x, Q2, f in zip(xs, Q2s, fs):
                f_out.append(dict(x=x, Q2=Q2, result=f))

        num_tab[obs_name] = f_out

    # remove QCDNUM cache files
    for f in [wname, zmname, hqname]:
        pathlib.Path(f).unlink(missing_ok=True)

    return num_tab



from .. import splitting_functions as split
from ..intrinsic import f3_cc as intrinsic
from ..light import f3_cc as light
from ..partonic_channel import RSL, EmptyPartonicChannel
from . import partonic_channel as pc


class AsyQuark(pc.PartonicChannelAsy, light.NonSingletEven):
    def NNLO(self):
        # silence NNLO since heavy NNLO still not available
        return RSL()


class AsyGluon(pc.PartonicChannelAsy):
    def NLO(self):
        def reg(z, args):
            return -self.L * split.lo.pqg_single(z, args)

        return RSL(reg)


class AsyLLNonSinglet(EmptyPartonicChannel):
    pass


class AsyNLLNonSinglet(EmptyPartonicChannel):
    pass


class AsyNNLLNonSinglet(EmptyPartonicChannel):
    pass


class AsyNNNLLNonSinglet(EmptyPartonicChannel):
    pass


class AsyLLIntrinsic(pc.PartonicChannelAsyLLIntrinsic):
    light_cls = light.NonSingletEven


class AsyNLLIntrinsicMatching(pc.PartonicChannelAsyNLLIntrinsicMatching):
    light_cls = light.NonSingletEven


class AsyNLLIntrinsicLight(pc.PartonicChannelAsyNLLIntrinsicLight):
    light_cls = light.NonSingletEven



import numpy as np

hqfl = "cbt"


def update(theory, observables):
    """
    Upgrade the legacy theory and observable runcards with the new settings.

    Parameters
    ----------
        theory : dict
            theory runcard
        observables : dict
            observable runcard

    Returns
    -------
        new_theory : dict
            upgraded theory runcard
        new_obs : dict
            upgraded observable runcard
    """
    new_theory = theory.copy()
    new_obs = observables.copy()
    update_fns(new_theory)
    update_scale_variations(new_theory)
    update_target(new_obs)
    if "alphaqed" in new_theory:
        new_theory["alphaem"] = new_theory.pop("alphaqed")
    # TODO: update Yadism syntax and remove PTO everywhere?
    if "QED" in new_theory:
        new_theory["order"] = (new_theory["PTO"] + 1, new_theory.pop("QED"))
    return new_theory, new_obs


def update_scale_variations(theory):
    """
    Extract necessity to compute scale variations.

    Parameters
    ----------
        theory : dict
            theory runcard
    """
    if "RenScaleVar" not in theory:
        theory["RenScaleVar"] = True
    if "FactScaleVar" not in theory:
        theory["FactScaleVar"] = True


def update_target(obs):
    """
    Map TargetDIS string to a (Z,A) dict.

    Parameters
    ----------
        obs : dict
            observable runcard
    """
    target = obs["TargetDIS"]
    if not isinstance(target, str):
        return
    if target == "proton":
        obs["TargetDIS"] = {"Z": 1.0, "A": 1.0}
        # proton = 2212
        obs["TargetDISid"] = "2212"
    elif target == "neutron":
        obs["TargetDIS"] = {"Z": 0.0, "A": 1.0}
        # neutron = 2112
        obs["TargetDISid"] = "2112"
    elif target == "isoscalar":
        obs["TargetDIS"] = {"Z": 1.0, "A": 2.0}
        # deuteron = 10 0 001 002 0
        obs["TargetDISid"] = "1000010020"
    elif target == "iron":
        obs["TargetDIS"] = {
            "Z": 23.403,
            "A": 49.618,
        }  # This target corresponds to NuTeV and has been determined through a
        # survey of the steel. See page 323 of the PhD thesis of David
        # Alexander Mason: https://inspirehep.net/files/15b639af43688d59f1ba65ec9cde7ec7
        # iron = 10 0 026 056 0
        obs["TargetDISid"] = "1000260560"
    elif target == "lead":
        obs["TargetDIS"] = {"Z": 82.0, "A": 208.0}  # from hep-ex/0102049
        # lead = 10 0 082 208 0
        obs["TargetDISid"] = "1000822080"
    elif target == "neon":
        obs["TargetDIS"] = {"Z": 10.0, "A": 20.0}
        obs["TargetDISid"] = "1000100200"
    elif target == "marble":
        A = (40 + 3 * 16 + 12) / 5  # A(CaCO3): average atomic mass
        Z = (20 + 3 * 8 + 6) / 5  # Z(CaCO3): avaerage atomic number
        obs["TargetDIS"] = {"Z": Z, "A": A}
        obs["TargetDISid"] = f"1000{str(Z)}{str(A)}0"
    else:
        raise ValueError(f"Unknown TargetDIS '{target}'")


def update_fns(theory):
    """
    Sets k{fl}Thr and  ZM{fl} for all heavy flavours depending on the scheme.

    Parameters
    ----------
        theory : dict
            theory runcard
    """
    fns = theory["FNS"]
    nf = theory["NfFF"]

    if fns == "ZM-VFNS":
        for fl in hqfl:
            theory[f"ZM{fl}"] = True
    elif fns == "FONLL-FFNS" or fns == "FONLL-FFN0":
        # enforce correct settings moving all thresholds to 0 or oo
        for k, fl in enumerate(hqfl):
            if k + 4 <= nf:
                theory[f"k{fl}Thr"] = 0.0
                theory[f"ZM{fl}"] = True
            elif k + 4 > nf + 1:
                theory[f"k{fl}Thr"] = np.inf
                theory[f"ZM{fl}"] = True
            else:
                # We only consider a single massive contribution. This is to
                # prevent double counting when different FNS are combined
                # to produce FONLL
                theory[f"k{fl}Thr"] = np.inf
                theory[f"ZM{fl}"] = False
    elif fns == "FFNS" or fns == "FFN0":
        # enforce correct settings moving all thresholds to 0 or oo
        for k, fl in enumerate(hqfl):
            if k + 4 <= nf:
                theory[f"k{fl}Thr"] = 0.0
                theory[f"ZM{fl}"] = True
            else:
                # for these flavours the massive contribution is calculated,
                # but they do not contribute to the number of running flavors
                theory[f"k{fl}Thr"] = np.inf
                theory[f"ZM{fl}"] = False
    else:
        raise ValueError(f"Scheme '{fns}' not recognized.")

    if "PTODIS" not in theory or theory["PTODIS"] is None:
        theory["PTODIS"] = theory["PTO"]

    if "FONLLParts" not in theory or theory["FONLLParts"] is None:
        theory["FONLLParts"] = "full"



"""High level interface to the observable names.

Wrap the name of the observables in order to perform operations on
them. For instance, it checks what kind of observable is being dealt
with and whether or not the given flavor is heavy.

"""

fake_kind = "??"
sfs = ["F2", "FL", "F3", "g1", "gL", "g4"]
# xs = ["XSreduced", "XSyreduced"]
xs = [
    "XSHERANC",
    "XSHERANCAVG",
    "XSHERACC",
    "XSCHORUSCC",
    "XSNUTEVCC",
    "XSNUTEVNU",
    "FW",
    "F1",
    "g5",
    "XSFPFCC",
]
kinds = sfs + xs + [fake_kind]
# external flavors:
heavys = ["charm", "bottom", "top"]
heavylights = [h + "light" for h in heavys]
external_flavors = heavys + ["light", "total"] + heavylights
# internally we allow in addition for the flavor families
flavors = external_flavors + ["heavy"]


class ObservableName:
    r"""Wrapper to observable names to easy split them into two parts.

    Parameters
    ----------
    name : str
        full observable name

    """

    def __init__(self, name):
        p = name.split("_")
        if len(p) == 1:
            self.kind, self.flavor = p[0], "total"
        elif len(p) == 2:
            self.kind, self.flavor = p[0], p[1]
        else:
            raise ValueError(f"Unknown obsname {name}")
        if self.kind not in kinds:
            raise ValueError(f"Unknown kind {self.kind}")
        if self.flavor not in flavors:
            raise ValueError(f"Unknown flavor {self.flavor}")

    @property
    def name(self):
        """Return joint name."""
        return self.kind + "_" + self.flavor

    @property
    def is_parity_violating(self):
        """Check if it is a parity violating observable."""
        if self.kind in ["F3", "gL", "g4"]:
            return True
        return False

    def __eq__(self, other):
        """Test equality of kind and flavor."""
        return self.kind == other.kind and self.flavor == other.flavor

    def apply_kind(self, kind):
        r"""Create new object with given kind and our flavor.

        Parameters
        ----------
        kind : str
            new kind

        Returns
        -------
        apply_kind : type(self)
            new kind and our flavor

        """
        return type(self)(kind + "_" + self.flavor)

    def __repr__(self):
        """Return the full representation name."""
        return self.name

    def apply_flavor(self, flavor):
        r"""Create new object with given flavor and our kind.

        Parameters
        ----------
        flavor : str
            new flavor

        Returns
        -------
        apply_flavor : type(self)
            our kind and new flavor

        """
        return type(self)(self.kind + "_" + flavor)

    @property
    def is_heavy(self):
        r"""Check if the given flavor is heavy.

        Returns
        -------
        is_heavy : bool
            is a heavy flavor?

        """
        return self.flavor != "light"

    @property
    def is_raw_heavy(self):
        """Check if it is a raw heavy flavor.

        i.e. charm, bottom, or, top.

        """
        return self.flavor in heavys

    @property
    def is_heavylight(self):
        """Check if it is a heavylight flavor.

        i.e. charmlight, bottomlight, or, toplight.

        """
        return self.flavor in heavylights

    @property
    def is_composed(self):
        """Check if it is a composed flavor.

        i.e. total.

        """
        return self.flavor == "total"

    @property
    def flavor_family(self):
        """Return abstract flavor family name."""
        if self.is_raw_heavy:
            return "heavy"
        if self.is_heavylight:
            return "light"
        return self.flavor

    def apply_flavor_family(self):
        r"""Return name with abstract flavor family name.

        Returns
        -------
        apply_flavor_family : type(self)
            new ObservableName

        """
        return self.apply_flavor(self.flavor_family)

    @property
    def hqnumber(self):
        """Return Heavy quark flavor number"""
        if self.is_heavylight:
            idx = heavylights.index(self.flavor)
        elif self.flavor_family in ["light", "total"]:
            idx = -4
        else:
            idx = heavys.index(self.flavor)
        return 4 + idx

    @property
    def raw_flavor(self):
        """Return underlying raw flavor."""
        if self.flavor == "light":
            return self.flavor
        return heavys[self.hqnumber - 4]

    @classmethod
    def has_heavies(cls, names):
        r"""Check if there are any heavy objects in names.

        Parameters
        ----------
        names : list(str)
            names to check

        Returns
        -------
        has_heavies : bool
            are there heavy obs in names?

        """
        for n in names:
            if not cls.is_valid(n):
                continue
            o = cls(n)
            if o.is_heavy:
                return True
        return False

    @classmethod
    def has_lights(cls, names):
        r"""Check if there are any light objects in names.

        Parameters
        ----------
        names : list(str)
            names to check

        Returns
        -------
        has_lights : bool
            are there light obs in names?

        """
        for n in names:
            if not cls.is_valid(n):
                continue
            o = cls(n)
            if o.flavor == "light":
                return True
        return False

    @property
    def mass_label(self):
        """Add mass label in the theory runcard."""
        if self.flavor == "light":
            return None
        else:
            return f"m{self.flavor[0]}"

    @classmethod
    def is_valid(cls, name):
        r"""Test whether the name is a valid observable name.

        Returns
        -------
        is_valid : bool
            is valid name?

        """
        try:
            cls(name)
            return True
        except ValueError:
            return False



"""Truely heavy structure function kernels."""

from eko import basis_rotation as br

from .. import kernels
from ..light.kernels import nc_weights as light_nc_weights


def import_pc_module(kind, process):
    """Import structure function submodule.

    Parameters
    ----------
    kind : str
        structure functions kind
    process : str
        DIS process type

    Returns
    -------
    module :
        suitable submodule
    """
    return kernels.import_local(kind, process, __name__)


def generate(esf, nf, ihq):
    """
    Collect the heavy coefficient functions.

    Parameters
    ----------
    esf : EvaluatedStructureFunction
        kinematic point
    nf : int
        number of light flavors
    ihq : int
        quark flavor to activate

    Returns
    -------
    elems : list(yadism.kernels.Kernel)
        list of elements

    """
    kind = esf.info.obs_name.kind
    is_pv = esf.info.obs_name.is_parity_violating
    pcs = import_pc_module(kind, esf.process)
    m2hq = esf.info.m2hq[ihq - 4]
    if esf.process == "CC":
        w = kernels.cc_weights(
            esf.info.coupling_constants,
            esf.Q2,
            br.quark_names[ihq - 1],
            nf,
            is_pv,
        )
        return (
            kernels.Kernel(w["ns"], pcs.NonSinglet(esf, nf, m2hq=m2hq)),
            kernels.Kernel(w["g"], pcs.Gluon(esf, nf, m2hq=m2hq)),
        )
    else:
        # F3 is a non-singlet quantity and hence has neither gluon nor singlet-like contributions
        if is_pv:
            return ()
        weights = nc_weights(
            esf.info.coupling_constants,
            esf.Q2,
            nf,
            ihq,
            is_pv,
        )
        n3lo_cf_variation = esf.info.theory["n3lo_cf_variation"]
        gVV = kernels.Kernel(
            weights["gVV"],
            pcs.GluonVV(esf, nf, m2hq=m2hq, n3lo_cf_variation=n3lo_cf_variation),
        )
        gAA = kernels.Kernel(weights["gAA"], pcs.GluonAA(esf, nf, m2hq=m2hq))
        sVV = kernels.Kernel(
            weights["sVV"],
            pcs.SingletVV(esf, nf, m2hq=m2hq, n3lo_cf_variation=n3lo_cf_variation),
        )
        sAA = kernels.Kernel(weights["sAA"], pcs.SingletAA(esf, nf, m2hq=m2hq))
        return (gVV, gAA, sVV, sAA)


def generate_missing(esf, nf, ihq, icoupl=None):
    """
    Collect the missing coefficient functions.

    Parameters
    ----------
    esf : EvaluatedStructureFunction
        kinematic point
    nf : int
        number of light flavors
    ihq : int
        PID of heavy flavor
    icoupl : None or int
        PID of the flavor coupling (default: None)

    Returns
    -------
    elems : list(yadism.kernels.Kernel)
        list of elements

    """
    # in CC there are no missing diagrams known yet
    if esf.process == "CC":
        return ()
    # only NC
    weights = light_nc_weights(
        esf.info.coupling_constants,
        esf.Q2,
        nf,
        esf.info.obs_name.is_parity_violating,
    )
    if icoupl is not None:
        weights["ns"] = {k: v for k, v in weights["ns"].items() if abs(k) == icoupl}
    kind = esf.info.obs_name.kind
    pcs = import_pc_module(kind, esf.process)
    m2hq = esf.info.m2hq[ihq - 4]
    return (kernels.Kernel(weights["ns"], pcs.NonSinglet(esf, nf, m2hq=m2hq)),)


def nc_weights(coupling_constants, Q2, nf, ihq, is_pv):
    """
    Compute heavy NC weights.

    Parameters
    ----------
    coupling_constants : CouplingConstants
        manager for coupling constants
    Q2 : float
        boson virtuality
    nf : int
        number of light flavors
    ihq : int
        quark flavor to activate
    is_pv: bool
        True if observable violates parity conservation

    Returns
    -------
    weights : dict
        mapping pid -> weight

    """
    if is_pv:
        return {}
    weight_vv = coupling_constants.get_weight(ihq, Q2, "VV")
    weight_aa = coupling_constants.get_weight(ihq, Q2, "AA")
    sVV = {}
    sAA = {}
    for q in range(1, nf + 1):
        sVV[q] = sVV[-q] = weight_vv
        sAA[q] = sAA[-q] = weight_aa
    return {"gVV": {21: weight_vv}, "gAA": {21: weight_aa}, "sVV": sVV, "sAA": sAA}



"""Coefficient function kernels.

Kernels are a combination of the coupling structure (which parton with which weight)
and the respective 'raw' mathematical coefficient function.
"""

import copy
import importlib
from numbers import Number

import numpy as np
from eko import basis_rotation as br


def import_local(kind, process, sibling):
    """Import the suitable subpackage with the actual partonic channel implementation.

    Parameters
    ----------
    kind : str
        structure function kind
    process : str
        DIS process type: "EM","NC","CC"
    sibling : str
        relative parent to import from

    Returns
    -------
    module : module
        module
    """
    kind = kind.lower()
    process = process.lower()
    if process == "em":
        process = "nc"
    parent = ".".join(sibling.split(".")[:-1])
    return importlib.import_module(f".{kind}_{process}", parent)


class Kernel:
    """Combination of partons with their weights and their mathematical expression in this channel.

    Parameters
    ----------
    partons : dict
        mapping pid -> weight
    coeff : PartonicChannel
        mathematical expression
    max_order : int
        if given, silence above this order
    min_order : int
        if given, silence below this order
    """

    def __init__(self, partons, coeff, max_order=None, min_order=None):
        self.partons = partons
        self.coeff = coeff
        self.max_order = max_order
        self.min_order = min_order

    def has_order(self, order):
        """Is current order active.

        Parameters
        ----------
        order : int
            order

        Returns
        -------
        bool :
            is active?
        """
        if self.min_order and order < self.min_order:
            return False
        if self.max_order and order > self.max_order:
            return False
        return True

    @property
    def channel(self):
        """Abstract classification in physical channels."""
        cls = str(type(self.coeff)).split("'")[1].split(".")[-1]

        # TODO: remove AsyQuark for AsyNonSinglet
        if "NonSinglet" in cls or "Quark" in cls:
            return "non-singlet"
        elif "Singlet" in cls:
            return "singlet"
        elif "Gluon" in cls:
            return "gluon"
        elif "Valence" in cls:
            return "valence"
        elif any(
            [x in cls for x in ["Intrinsic", "Splus", "Sminus", "Rplus", "Rminus"]]
        ):
            return "intrinsic"
        else:
            raise ValueError(f"Class '{cls}' does not correspond to a known channel")

    def __repr__(self):
        return repr({"partons": self.partons, "coeff": self.coeff})

    def __neg__(self):
        return self.__rmul__(-1)

    def __mul__(self, f):
        return self.__rmul__(f)

    def __rmul__(self, f):
        if not isinstance(f, Number):
            raise ValueError("Can only multiply numbers")
        partons = {k: f * v for k, v in self.partons.items()}
        return self.__class__(
            partons, copy.copy(self.coeff), self.max_order, self.min_order
        )


def cc_weights(coupling_constants, Q2, cc_mask, nf, is_pv):
    """Collect the weights of the partons.

    Parameters
    ----------
    coupling_constants : CouplingConstants
        manager for coupling constants
    Q2 : float
        W virtuality
    cc_mask : str
        participating flavors on the CKM matrix
    nf : int
        number of light flavors
    is_pv: bool
        True if observable violates parity conservation

    Returns
    -------
    weights : dict
        mapping pid -> weight for q and g channel
    """
    weights = {"ns": {}, "g": {}, "s": {}}
    # determine couplings
    projectile_pid = coupling_constants.obs_config["projectilePID"]
    if projectile_pid in [-11, 12]:
        rest = 1
    else:
        rest = 0
    # quark couplings
    tot_ch_sq = 0
    norm = len(cc_mask)
    # iterate: include the heavy quark itself, since it can run in the singlet sector diagrams
    for q in range(1, min(nf + 2, 6 + 1)):
        sign = 1 if q % 2 == rest else -1
        w = coupling_constants.get_weight(q, Q2, None, cc_mask=cc_mask)
        # the heavy quark can not be in the input
        # NOTE: intrinsic abuse this statement with nf -> nf + 1
        if q <= nf:
            # @F3-sign@
            weights["ns"][sign * q] = w if not is_pv else sign * w
        # but it contributes to the average
        tot_ch_sq += w
    # gluon coupling = charge sum
    if rest == 0 and is_pv:
        tot_ch_sq *= -1
    weights["g"][21] = tot_ch_sq / norm / 2
    # add singlet
    for q in weights["ns"]:
        weights["s"][q] = tot_ch_sq / norm / 2
        weights["s"][-q] = tot_ch_sq / norm / 2
    return weights


def cc_weights_even(coupling_constants, Q2, cc_mask, nf, is_pv):
    """Collect the weights of the partons.

    Parameters
    ----------
    coupling_constants : CouplingConstants
        manager for coupling constants
    Q2 : float
        W virtuality
    cc_mask : str
        participating flavors on the CKM matrix
    nf : int
        number of light flavors
    is_pv: bool
        True if observable violates parity conservation

    Returns
    -------
    weights : dict
        mapping pid -> weight for q and g channel
    """
    weights = {"ns": {}, "g": {}, "s": {}}
    # determine couplings
    projectile_pid = coupling_constants.obs_config["projectilePID"]
    if projectile_pid in [-11, 12]:
        rest = 1
    else:
        rest = 0
    # quark couplings
    tot_ch_sq = 0
    norm = len(cc_mask)
    # iterate: include the heavy quark itself, since it can run in the singlet sector diagrams
    for q in range(1, min(nf + 2, 6 + 1)):
        sign = 1 if q % 2 == rest else -1
        w = coupling_constants.get_weight(q, Q2, None, cc_mask=cc_mask)
        # the heavy quark can not be in the input
        # NOTE: intrinsic abuse this statement with nf -> nf + 1
        if q <= nf:
            # @F3-sign@
            weights["ns"][sign * q] = w / 2 * (1 if not is_pv else sign)
            weights["ns"][-sign * q] = w / 2 * (1 if not is_pv else sign)
        # but it contributes to the average
        tot_ch_sq += w
    # gluon coupling = charge sum
    weights["g"][21] = tot_ch_sq / norm / 2
    # add singlet
    for q in range(1, nf + 1):
        weights["s"][q] = tot_ch_sq / norm / 2
        weights["s"][-q] = tot_ch_sq / norm / 2
    return weights


def cc_weights_odd(coupling_constants, Q2, cc_mask, nf, is_pv):
    """Collect the weights of the partons.

    Parameters
    ----------
    coupling_constants : CouplingConstants
        manager for coupling constants
    Q2 : float
        W virtuality
    cc_mask : str
        participating flavors on the CKM matrix
    nf : int
        number of light flavors
    is_pv: bool
        True if observable violates parity conservation

    Returns
    -------
    weights : dict
        mapping pid -> weight for q and g channel
    """
    weights = {"ns": {}, "v": {}}
    # determine couplings
    projectile_pid = coupling_constants.obs_config["projectilePID"]
    if projectile_pid in [-11, 12]:
        rest = 1
    else:
        rest = 0
    # quark couplings
    tot_ch_sq = 0
    norm = len(cc_mask)
    # iterate: include the heavy quark itself, since it can run in the singlet sector diagrams
    for q in range(1, min(nf + 2, 6 + 1)):
        sign = 1 if q % 2 == rest else -1
        w = coupling_constants.get_weight(q, Q2, None, cc_mask=cc_mask)
        # the heavy quark can not be in the input
        # NOTE: intrinsic abuse this statement with nf -> nf + 1
        if q <= nf:
            # @F3-sign@
            weights["ns"][sign * q] = w / 2 * (1 if not is_pv else sign)
            weights["ns"][-sign * q] = -w / 2 * (1 if not is_pv else sign)
        tot_ch_sq += w
    # add valence, here there is no need to distinguish up and down.
    for q in range(1, nf + 1):
        weights["v"][q] = tot_ch_sq / norm / 2
        weights["v"][-q] = -tot_ch_sq / norm / 2
    return weights


def generate_single_flavor_light(esf, nf, ihq):
    """Add a light-like contribution for a single quark flavor.

    The linear dependency to the electric charge is introduce by multiplying
    and diving by nf. The multiplication is *implicit* inside the coefficient function,
    the division is *explicit* made here.

    Parameters
    ----------
    esf : EvaluatedStructureFunction
        kinematic point
    nf : int
        number of light flavors
    ihq : int
        quark flavor to activate

    Returns
    -------
    elems : list(yadism.kernels.Kernel)
        list of elements
    """
    kind = esf.info.obs_name.kind
    is_pv = esf.info.obs_name.is_parity_violating
    light_cfs = import_local(
        kind, esf.process, ".".join(__name__.split(".")[:-1] + ["light", ""])
    )
    if esf.process == "CC":
        w_even = cc_weights_even(
            esf.info.coupling_constants,
            esf.Q2,
            br.quark_names[ihq - 1],
            nf,
            is_pv,
        )
        ns_even = Kernel(w_even["ns"], light_cfs.NonSingletEven(esf, nf))
        w_odd = cc_weights_odd(
            esf.info.coupling_constants,
            esf.Q2,
            br.quark_names[ihq - 1],
            nf,
            is_pv,
        )
        ns_odd = Kernel(w_odd["ns"], light_cfs.NonSingletOdd(esf, nf))

        if is_pv:
            v = Kernel(
                {k: np.sign(k) * c / (nf) for k, c in w_odd["v"].items()},
                light_cfs.Valence(esf, nf),
            )
            return (ns_even, ns_odd, v)
        g = Kernel({21: w_even["g"][21] / (nf)}, light_cfs.Gluon(esf, nf))
        s = Kernel(
            {k: v / (nf) for k, v in w_even["s"].items()}, light_cfs.Singlet(esf, nf)
        )
        return (ns_even, g, s, ns_odd)

    # NC
    if is_pv:
        w = esf.info.coupling_constants.get_weight(
            ihq, esf.Q2, "VA"
        ) + esf.info.coupling_constants.get_weight(ihq, esf.Q2, "AV")
    else:
        w = esf.info.coupling_constants.get_weight(
            ihq, esf.Q2, "VV"
        ) + esf.info.coupling_constants.get_weight(ihq, esf.Q2, "AA")

    ns_partons = {}
    ns_partons[ihq] = w
    ns_partons[-ihq] = w if not is_pv else -w
    ch_av = w / nf  # omitting again *2/2

    # all quarks from 1 up to and including nf for the singlet and valence
    pids = range(1, nf + 1)
    if is_pv:
        v_partons = {q: np.sign(q) * ch_av for q in [*pids, *(-q for q in pids)]}
        return (
            Kernel(ns_partons, light_cfs.NonSinglet(esf, nf)),
            Kernel(v_partons, light_cfs.Valence(esf, nf)),
        )

    s_partons = {q: ch_av for q in [*pids, *(-q for q in pids)]}
    kernels_list = [
        Kernel(ns_partons, light_cfs.NonSinglet(esf, nf)),
        Kernel({21: ch_av}, light_cfs.Gluon(esf, nf)),
        Kernel(s_partons, light_cfs.Singlet(esf, nf)),
    ]

    if esf.info.theory["pto"] == 3:
        w = esf.info.coupling_constants.get_fl11_weight(
            ihq, esf.Q2, nf, "VV"
        ) + esf.info.coupling_constants.get_fl11_weight(ihq, esf.Q2, nf, "AA")
        quark_partons = {}
        quark_partons[ihq] = w
        quark_partons[-ihq] = w
        ch_av = w / nf
        kernels_list.extend(
            [
                Kernel(quark_partons, light_cfs.QuarkFL11(esf, nf)),
                Kernel({21: ch_av}, light_cfs.GluonFL11(esf, nf)),
            ]
        )
    return kernels_list



import numba as nb
import numpy as np
from eko import constants

from ... import special
from . import non_singlet as ns


@nb.njit("f8(f8,f8[:])", cache=True)
def pgq0_reg(z, _args):
    """
    (Regular) :math:`P_{gq}` splitting function.

    |ref| implements :eqref:`4.6`, :cite:`split-singlet`.

    Parameters
    ----------
        z : float
            partonic momentum fraction
        args : np.ndarray
            further arguments

    Returns
    -------
        float
            the gluon-quark splitting function :math:`P_{gq}(z)`

    """
    return 2.0 * constants.CF * (2 / z - 2 + z)


@nb.njit("f8(f8,f8[:])", cache=True)
def pgg0_reg(z, _args):
    """
    Regular :math:`P_{gg}` splitting function.

    |ref| implements :eqref:`4.6`, :cite:`split-singlet`.

    Parameters
    ----------
        z : float
            partonic momentum fraction
        args : np.ndarray
            further arguments

    Returns
    -------
        float
            the gluon-gluon splitting function :math:`P_{gg}(z)`

    """
    return 4 * constants.CA * (1 / z - 2 + z - z**2)


@nb.njit("f8(f8,f8[:])", cache=True)
def pgg0_sing(z, _args):
    """
    Singular :math:`P_{gg}` splitting function.

    |ref| implements :eqref:`4.6`, :cite:`split-singlet`.

    Parameters
    ----------
        z : float
            partonic momentum fraction
        args : np.ndarray
            further arguments

    Returns
    -------
        float
            the gluon-gluon splitting function :math:`P_{gg}(z)`

    """
    return 4 * constants.CA / (1 - z)


@nb.njit("f8(f8,f8[:])", cache=True)
def pgg0_local(x, args):
    """
    Singular :math:`P_{gg}` splitting function.

    |ref| implements :eqref:`4.6`, :cite:`split-singlet`.

    Parameters
    ----------
        z : float
            partonic momentum fraction
        args : np.ndarray
            further arguments

    Returns
    -------
        float
            the gluon-gluon splitting function :math:`P_{gg}(z)`

    """
    return constants.CA * (4 * np.log(1 - x) + 11 / 3) - args[0] * 2 / 3


@nb.njit("f8(f8,f8[:])", cache=True)
def pqq1ps_reg(z, args):
    """
    Regular :math:`P_{qq}` pure singlet splitting function.

    Parameters
    ----------
        z : float
            partonic momentum fraction
        args : np.ndarray
            further arguments

    Returns
    -------
        float
            the quark-quark pure singlet splitting function :math:`P_{qq}(z)`

    """
    nf = args[0]
    x = z

    DX = 1 / x
    LNX = np.log(x)
    HR200 = LNX * LNX / 2

    #  The splitting function in terms of the harmonic polylogs
    X1PSA = (
        +nf
        * constants.CF
        * (
            -8
            + 24 * x
            - 224 / 9 * x**2
            + 80 / 9 * DX
            + 4 * LNX
            + 20 * LNX * x
            + 32 / 3 * LNX * x**2
            - 8 * HR200
            - 8 * HR200 * x
        )
    )

    return X1PSA


@nb.njit("f8(f8,f8[:])", cache=True)
def pqq1_reg(z, args):
    """
    Regular :math:`P_{qq}` splitting function.

    Parameters
    ----------
        z : float
            partonic momentum fraction
        args : np.ndarray
            further arguments

    Returns
    -------
        float
            the quark-quark singlet splitting function :math:`P_{qq}(z)`

    """
    return pqq1ps_reg(z, args) + ns.pnsp_reg(z, args)


@nb.njit("f8(f8,f8[:])", cache=True)
def pqg1_reg(z, args):
    """
    (Regular) :math:`P_{qg}` splitting function.

    |ref| implements :eqref:`4.6`, :cite:`split-singlet`.

    Parameters
    ----------
        z : float
            partonic momentum fraction
        args : np.ndarray
            further arguments

    Returns
    -------
        float
            the quark-gluon splitting function :math:`P_{qg}(z)`

    """
    CA = constants.CA
    CF = constants.CF
    NF = args[0]
    x = z

    lnx = np.log(x)
    ln1mx = np.log(1 - x)
    pqg = x**2 + (1 - x) ** 2
    pqgmx = x**2 + (1 + x) ** 2
    S2x = special.s2(x)

    X1QGA = 2 * CF * NF * (
        4
        + 4 * ln1mx
        + (10 - 4 * (ln1mx - lnx) + 2 * (-ln1mx + lnx) ** 2 - 2 * np.pi**2 / 3) * pqg
        - lnx * (1 - 4 * x)
        - lnx**2 * (1 - 2 * x)
        - 9 * x
    ) + 2 * CA * NF * (
        20.22222222222222
        - 4 * ln1mx
        + (
            -24.22222222222222
            + 4 * ln1mx
            - 2 * ln1mx**2
            + (44 * lnx) / 3
            - lnx**2
            + np.pi**2 / 3
        )
        * pqg
        + 2 * pqgmx * S2x
        + 40 / (9 * x)
        + (14 * x) / 9
        - lnx**2 * (2 + 8 * x)
        + lnx * (-12.666666666666666 + (136 * x) / 3)
    )
    return X1QGA



"""Yet Another DIS Module."""

from . import version
from .runner import Runner

__version__ = version.__version__


def run_yadism(theory: dict, observables: dict):
    r"""Call yadism runner.

    Get the theory and observables description and computes the
    Coefficient Functions for the requested kinematics.

    Parameters
    ----------
    theory: dict
        Dictionary containing the theory Parameters
    observables: dict
        Dictionary containing the DIS parameters such as
        process description and kinematic specifications

    Returns
    -------
    run_yadism: :obj:`Output`
        output object containing the coefficient functions
        grids

    """
    runner = Runner(theory, observables)
    return runner.get_result()



from ..light import f3_nc as light
from ..partonic_channel import EmptyPartonicChannel
from . import g1_nc
from . import partonic_channel as pc


# NOTE in the non-singlet case d_{2xg1} = d_{xF3}
class AsyLLNonSinglet(g1_nc.AsyLLNonSinglet):
    pass


class AsyNLLNonSinglet(g1_nc.AsyNLLNonSinglet):
    pass


class AsyNNLLNonSinglet(g1_nc.AsyNNLLNonSinglet):
    pass


class AsyNNNLLNonSinglet(EmptyPartonicChannel):
    pass


class AsyLLIntrinsic(pc.PartonicChannelAsyLLIntrinsic):
    light_cls = light.NonSinglet


class AsyNLLIntrinsicMatching(pc.PartonicChannelAsyNLLIntrinsicMatching):
    light_cls = light.NonSinglet


class AsyNLLIntrinsicLight(pc.PartonicChannelAsyNLLIntrinsicLight):
    light_cls = light.NonSinglet
